!
! Module: binned_att
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This module contains routines associated with binned-up attitude data.
! 
! Comments: 
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module binned_att

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the '*' from the lib*.a name
! in <package>/src/Makefile.

  use dal
  use errorHandling
  use caloaldefs	! for SpacecraftAttitudeType
  use caltypes		! for IntervalsT
  use oal

  use array_utils	! for getBinNumber
  use constants
  use dal_aux
  use intervals_aux	! for gtiVecType, andIntervals.

  implicit none 

  private 

  public initializeBinnedAtt
  public binUpAttitudeFromOdf
  public sampleBinnedAttitude
  public binUpAttitude
  public writeBinnedAttitude
  public readBinnedAttitude
  public gtiWeightBinnedAttitude
  public calcBinnedAttGtis
  public openGtiWeightBinnedAtt
  public constructDummyBinnedAtt

  type, public :: attBinType
    type(SpacecraftAttitudeType) ::&
      att
    real(double) ::&
      binEndTime=0.0d0
    real(single) ::&
      fracExp=0.0
    logical ::&
      isGood=.false.
  end type attBinType

contains 

  !---------------------------------------------------------------------
  subroutine initializeBinnedAtt(binnedAtt)
  !---------------------------------------------------------------------

    implicit none

    type(attBinType), intent(out) :: binnedAtt(:)

    !-------------------------------------------------------------------
    ! Start

    binnedAtt%att%ra     = 0.0d0
    binnedAtt%att%dec    = 0.0d0
    binnedAtt%att%apos   = 0.0d0
    binnedAtt%binEndTime = 0.0d0
    binnedAtt%fracExp    = 0.0
    binnedAtt%isGood     = .false.

  end subroutine initializeBinnedAtt

  !---------------------------------------------------------------------
  subroutine binUpAttitudeFromOdf(time, maxDelta, attInBin, attIsGood&
  , attIsNew)
  !---------------------------------------------------------------------
  ! The subroutine samples the attitude from the AHF in the ODF. If the
  ! attitude is good, the returned value is the last good attitude unless
  ! (i) this is the first time a good attitude has been returned from the
  ! AHF, or (ii) the difference between the present attitude and the last
  ! good one exceeds certain limits; in either of these cases, the
  ! attitude returned by the routine is that just obtained from the AHF.
  !
  ! NOTE! The OAL should be set up for attitudeFromAhf before this sub
  ! is called.

    implicit none 

    real(double),                 intent(in)  :: time
    type(SpacecraftAttitudeType), intent(in)  :: maxDelta
    type(SpacecraftAttitudeType), intent(out) :: attInBin
    logical,                      intent(out) :: attIsGood, attIsNew

    type(SpacecraftAttitudeType) :: att
    logical, save :: noGoodAttsYet = .true., lastAttWasGood = .false.
    type(SpacecraftAttitudeType), save :: baselineAtt
    logical :: raIsNew, decIsNew, paIsNew

    !-------------------------------------------------------------------
    ! Start:

    attInBin%ra   = 0.0d0
    attInBin%dec  = 0.0d0
    attInBin%apos = 0.0d0

    if (OAL_getAttitude(time, att)) then 
      attIsGood = .true.
    else
      attIsGood = .false.
    end if

!attIsGood = .true.
!att%ra = 1.5707963268d0
!att%dec = 0.0d0
!att%apos = 0.0d0
!print *, 'CAREFUL! Temporary test routine activated.'

    attIsNew = .false. ! default

    if (attIsGood) then
      if (noGoodAttsYet) then
        attInBin = att
        baselineAtt = att

      else ! some good atts have already been found, thus guaranteeing
      ! the existence of a baseline to test against.
        raIsNew  = .false.
        decIsNew = .false.
        paIsNew  = .false.

        if (abs(att%ra - baselineAtt%ra) > maxDelta%ra) then
!***** NOTE! At present this does NOT cater for situations in which the attitude wanders across ra=0/pi, dec=+-90 or apos=0/pi. Better would be to work in a pnt-centric coordinate system.

          baselineAtt%ra = att%ra
          raIsNew = .true.
        endif
        if (abs(att%dec - baselineAtt%dec) > maxDelta%dec) then 
          baselineAtt%dec = att%dec
          decIsNew = .true.
        endif
        if (abs(att%apos - baselineAtt%apos) > maxDelta%apos) then 
          baselineAtt%apos = att%apos
          paIsNew = .true.
        endif

        if (raIsNew .or. decIsNew .or. paIsNew) then 
          attInBin = att
          baselineAtt = att
          attIsNew = .true.
        else
          attInBin = baselineAtt
        end if
      end if

      if (.not. lastAttWasGood) attIsNew = .true.
      noGoodAttsYet = .false.
      lastAttWasGood = .true.

    else ! att is not good
      if (lastAttWasGood) then
        attIsNew = .true.
      else ! last att was no good
      end if
      lastAttWasGood = .false.
    end if

  end subroutine binUpAttitudeFromOdf

  !---------------------------------------------------------------------
  subroutine sampleBinnedAttitude(binnedAtt, startTimeFirstBin, time&
  , endsUnbounded, scAttitude, attIsGood, attIsNew)
  !---------------------------------------------------------------------

    implicit none

    type(attBinType),             intent(in)  :: binnedAtt(:)
    real(double),                 intent(in)  :: startTimeFirstBin, time
    logical,                      intent(in)  :: endsUnbounded(2)
    type(SpacecraftAttitudeType), intent(out) :: scAttitude
    logical,                      intent(out) :: attIsGood, attIsNew

    integer :: numAttBins, binNum
    logical, save :: noGoodAttsYet = .true.
    integer, save :: binOfLastGoodAtt = 0
!real(double), allocatable :: timeBinEdges(:)

    !-------------------------------------------------------------------
    ! Start

    scAttitude%ra   = 0.0d0
    scAttitude%dec  = 0.0d0
    scAttitude%apos = 0.0d0
    attIsGood = .false.
    attIsNew  = .false.

    numAttBins = size(binnedAtt)
!allocate(timeBinEdges(0:numAttBins))
!timeBinEdges(0) = startTimeFirstBin
!timeBinEdges(1:) = binnedAtt%binEndTime

    if (time < startTimeFirstBin) then
      binNum = 1

      if (endsUnbounded(1)) then
        if (binnedAtt(binNum)%isGood) then
          attIsGood = .true.
          scAttitude = binnedAtt(binNum)%att
        else
          attIsGood = .false.
        end if
      else
        attIsGood = .false.
      end if

    else if (time > binnedAtt(numAttBins)%binEndTime) then
      binNum = numAttBins

      if (endsUnbounded(2)) then
        if (binnedAtt(binNum)%isGood) then
          attIsGood = .true.
          scAttitude = binnedAtt(binNum)%att
        else
          attIsGood = .false.
        end if
      else
        attIsGood = .false.
      end if

    else
      binNum = getBinNumberDouble((/startTimeFirstBin, binnedAtt%binEndTime/)&
!binNum = getBinNumber(timeBinEdges&
        , time) ! in ssclib/array_utils.

      if (binnedAtt(binNum)%isGood) then
        attIsGood = .true.
        scAttitude = binnedAtt(binNum)%att
      else
        attIsGood = .false.
      end if
    end if

    if (attIsGood) then
      if (noGoodAttsYet) then
        noGoodAttsYet = .false.
        binOfLastGoodAtt = binNum
        attIsNew = .true.

      else ! we can be sure that binOfLastGoodAtt is defined.
        if (binNum == binOfLastGoodAtt) then
          attIsNew = .false.
        else
          binOfLastGoodAtt = binNum
          attIsNew = .true.
        end if
      end if
    end if

  end subroutine sampleBinnedAttitude

  !---------------------------------------------------------------------
  function getBinNumberDouble(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

!***** NOTE: should be in ssclib>3.16.

    implicit none

    real(double), intent(in) :: xBinEdges(:), x
    integer(int32)           :: binNum

    !-------------------------------------------------------------------
    ! Start

    if (size(xBinEdges) < 2) then
      binNum = 0
  return
    end if

    binNum = largestTrueIndex(logical(xBinEdges <= x, bool))

  end function getBinNumberDouble

  !---------------------------------------------------------------------
  subroutine binUpAttitude()
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

!*****

  end subroutine binUpAttitude

  !---------------------------------------------------------------------
  subroutine writeBinnedAttitude(binnedAttSetName, binnedAtt&
  , startTimeFirstBin, pntDeg, maxDelta, pntStyle, endsUnbounded)
  !---------------------------------------------------------------------

    implicit none

    character(*),     intent(in) :: binnedAttSetName
    type(attBinType), intent(in) :: binnedAtt(:)
    real(double),     intent(in) :: startTimeFirstBin
    type(SpacecraftAttitudeType),&
                      intent(in) :: pntDeg, maxDelta
    character(*),     intent(in) :: pntStyle
    logical,          intent(in) :: endsUnbounded(2)

    type(DataSetT) :: binnedAttSet
    type(TableT) :: binnedAttTab
    real(double), pointer :: time(:), raDeg(:), decDeg(:), paDeg(:)
    real(single), pointer :: weight(:)
    logical(bool), pointer :: isGood(:)

    !-------------------------------------------------------------------
    ! Start

    binnedAttSet = dataSet(binnedAttSetName, CREATE)
    binnedAttTab = addTable(binnedAttSet, 'ATT_BINS', size(binnedAtt))

    call setAttribute(binnedAttTab, 'TIMEZERO', startTimeFirstBin)
    call setAttribute(binnedAttTab, 'STARTINF', endsUnbounded(1))
    call setAttribute(binnedAttTab, 'END__INF', endsUnbounded(2))
    call setAttribute(binnedAttTab, 'RA_PNT',   pntDeg%ra,"deg"&
      ,"Average right ascension of the s/c pointing axis")
    call setAttribute(binnedAttTab, 'DEC_PNT',  pntDeg%dec,"deg"&
      ,"Average declination of the s/c pointing axis")
    call setAttribute(binnedAttTab, 'PA_PNT',   pntDeg%apos,"deg"&
      ,"Average position angle of the s/c pointing axis")
    call setAttribute(binnedAttTab, 'RA_DELTA', maxDelta%ra   / PiOn180)
    call setAttribute(binnedAttTab, 'DECDELTA', maxDelta%dec  / PiOn180)
    call setAttribute(binnedAttTab, 'PA_DELTA', maxDelta%apos / PiOn180)

    call setAttribute(binnedAttTab, 'AVRG_PNT', trim(pntStyle)&
      ,"Meaning of _PNT values (mean, median or fixed)")

    time   => real64Data(addColumn(binnedAttTab, 'TSTOP',   REAL64, 's'))
    raDeg  => real64Data(addColumn(binnedAttTab, 'RA',      REAL64, 'deg'))
    decDeg => real64Data(addColumn(binnedAttTab, 'DEC',     REAL64, 'deg'))
    paDeg  => real64Data(addColumn(binnedAttTab, 'PA',      REAL64, 'deg'))
    weight => real32Data(addColumn(binnedAttTab, 'FRACEXP', REAL32))
    isGood => boolData(  addColumn(binnedAttTab, 'IS_GOOD', BOOLEAN))

    time   = binnedAtt%binEndTime
    raDeg  = binnedAtt%att%ra   / PiOn180
    decDeg = binnedAtt%att%dec  / PiOn180
    paDeg  = binnedAtt%att%apos / PiOn180
    weight = binnedAtt%fracexp
    isGood = binnedAtt%isGood

    call release(binnedAttSet)

!****** also write HDU and other OGIP keywords????

  end subroutine writeBinnedAttitude

  !---------------------------------------------------------------------
  subroutine readBinnedAttitude(binnedAttSetName, binnedAtt&
  , startTimeFirstBin, endsUnbounded)
  !---------------------------------------------------------------------
  ! Note! The calling routine should deallocate the pointer binnedAtt after use.

    implicit none

    character(*),     intent(in)  :: binnedAttSetName
    type(attBinType), pointer     :: binnedAtt(:)
    real(double),     intent(out) :: startTimeFirstBin
    logical,          intent(out) :: endsUnbounded(2)

    type(DataSetT) :: binnedAttSet
    type(TableT) :: binnedAttTab
    integer :: numAttBins!, attBinNum
    logical(bool), allocatable :: attIsGoodBool(:)

    !-------------------------------------------------------------------
    ! Start:

    binnedAttSet = dataset(binnedAttSetName, READ)
    binnedAttTab = table(binnedAttSet, 'ATT_BINS')

    numAttBins = numberOfRows(binnedAttTab)
    allocate(binnedAtt(numAttBins))
    allocate(attIsGoodBool(numAttBins))

    startTimeFirstBin = real64Attribute(binnedAttTab, 'TIMEZERO')
    endsUnbounded(1) = booleanAttribute(binnedAttTab, 'STARTINF')
    endsUnbounded(2) = booleanAttribute(binnedAttTab, 'END__INF')

    call readColDataToFixed(binnedAttTab, 'TSTOP',   binnedAtt%binEndTime)
    call readColDataToFixed(binnedAttTab, 'RA',      binnedAtt%att%ra)
    call readColDataToFixed(binnedAttTab, 'DEC',     binnedAtt%att%dec)
    call readColDataToFixed(binnedAttTab, 'PA',      binnedAtt%att%apos)
    call readColDataToFixed(binnedAttTab, 'FRACEXP', binnedAtt%fracexp)
    call readColDataToFixed(binnedAttTab, 'IS_GOOD', attIsGoodBool)
    ! These calls are in ssclib/dal_aux.

    binnedAtt%att%ra   = binnedAtt%att%ra   * PiOn180
    binnedAtt%att%dec  = binnedAtt%att%dec  * PiOn180
    binnedAtt%att%apos = binnedAtt%att%apos * PiOn180
    binnedAtt%isGood   = attIsGoodBool

    call release(binnedAttSet)

  end subroutine readBinnedAttitude

  !---------------------------------------------------------------------
  subroutine gtiWeightBinnedAttitude(binnedAtt, startTimeFirstBin&
    , endsUnbounded, gtis, multiplyByGtis, binnedAttWeights)
  !---------------------------------------------------------------------

    implicit none

    type(attBinType), intent(in)  :: binnedAtt(:)
    real(double),     intent(in)  :: startTimeFirstBin
    logical,          intent(in)  :: endsUnbounded(2)
    type(IntervalT),  intent(in)  :: gtis(:)
    logical,          intent(in)  :: multiplyByGtis
    real(single),     intent(out) :: binnedAttWeights(size(binnedAtt))

    integer :: numAttBins, numGtis, attBinNum
    type(IntervalT), allocatable :: attIntervals(:)
    type(IntervalT), pointer :: andedIntervals(:)

    !-------------------------------------------------------------------
    ! Start

    numAttBins = size(binnedAtt)
  if (numAttBins <= 0) return

    numGtis = size(gtis)
    if (numGtis <= 0) then
      binnedAttWeights = 0.0
  return
    end if

    allocate(attIntervals(numAttBins))
    attIntervals(1)%lower = startTimeFirstBin
    attIntervals(1)%upper = binnedAtt(1)%binEndTime
    do attBinNum = 2, numAttBins
      attIntervals(attBinNum)%lower = attIntervals(attBinNum-1)%upper
      attIntervals(attBinNum)%upper = binnedAtt(attBinNum)%binEndTime
    end do

    if (endsUnbounded(1)) then
      if (endsUnbounded(2)) then
        attIntervals(1         )%lower = gtis(1      )%lower
        attIntervals(numAttBins)%upper = gtis(numGtis)%upper

        if (numAttBins > 1) then
          attIntervals(1         )%lower = min(attIntervals(1         )%lower&
            , attIntervals(1         )%upper)
          attIntervals(numAttBins)%upper = max(attIntervals(numAttBins)%lower&
            , attIntervals(numAttBins)%upper)
        end if

      else
        attIntervals(1)%lower = min(gtis(1)%lower, attIntervals(1)%upper)
      end if

    else
      if (endsUnbounded(2)) then
        attIntervals(numAttBins)%upper = max(gtis(numGtis)%upper&
          , attIntervals(numAttBins)%lower)
      end if
    end if

    do attBinNum = 1, numAttBins
      call andIntervals(attIntervals(attBinNum), gtis, andedIntervals)
      ! in ssclib/intervals_aux

      binnedAttWeights(attBinNum) = sum(andedIntervals%upper)&
         - sum(andedIntervals%lower)

      deallocate(andedIntervals)
    end do

    if (.not. multiplyByGtis) binnedAttWeights = binnedAttWeights&
      / (attIntervals(numAttBins)%upper - attIntervals(1)%lower)

  end subroutine gtiWeightBinnedAttitude

  !---------------------------------------------------------------------
  subroutine calcBinnedAttGtis(binnedAtt, startTimeFirstBin&
    , endsUnbounded, gtis, gtiType, binnedAttWeights)
  !---------------------------------------------------------------------

    implicit none

    type(attBinType), intent(in)  :: binnedAtt(:)
    real(double),     intent(in)  :: startTimeFirstBin
    logical,          intent(in)  :: endsUnbounded(2)
    type(IntervalT),  intent(in)  :: gtis(:)
    integer,          intent(in)  :: gtiType
    real(single),     intent(out) :: binnedAttWeights(size(binnedAtt))

    integer :: numAttBins, numGtis, attBinNum
    type(IntervalT), allocatable :: attIntervals(:)
    type(IntervalT), pointer :: andedIntervals(:)

!    integer, parameter :: OPEN_GTI=0, CLOSED_GTI=1, GTI_IN_LIST=2
!*** Yuk. Better to move the whole gtiVecType to here (or to intervals_aux).

    !-------------------------------------------------------------------
    ! Start

    numAttBins = size(binnedAtt)
  if (numAttBins <= 0) return

    numGtis = size(gtis)
    if (gtiType == GTI_IN_LIST .and. numGtis <= 0) then
      binnedAttWeights = 0.0
  return
    end if

    if (gtiType == CLOSED_GTI) then
      binnedAttWeights = 0.0
  return
    end if

    if (any(endsUnbounded) .and. gtiType == OPEN_GTI)&
call error('',"")

    allocate(attIntervals(numAttBins))
    attIntervals(1)%lower = startTimeFirstBin
    attIntervals(1)%upper = binnedAtt(1)%binEndTime
    do attBinNum = 2, numAttBins
      attIntervals(attBinNum)%lower = attIntervals(attBinNum-1)%upper
      attIntervals(attBinNum)%upper = binnedAtt(attBinNum)%binEndTime
    end do

    if (endsUnbounded(1)) then
      if (endsUnbounded(2)) then
        attIntervals(1         )%lower = gtis(1      )%lower
        attIntervals(numAttBins)%upper = gtis(numGtis)%upper

        if (numAttBins > 1) then
          attIntervals(1         )%lower = min(attIntervals(1         )%lower&
            , attIntervals(1         )%upper)
          attIntervals(numAttBins)%upper = max(attIntervals(numAttBins)%lower&
            , attIntervals(numAttBins)%upper)
        end if

      else
        attIntervals(1)%lower = min(gtis(1)%lower, attIntervals(1)%upper)
      end if

    else
      if (endsUnbounded(2)) then
        attIntervals(numAttBins)%upper = max(gtis(numGtis)%upper&
          , attIntervals(numAttBins)%lower)
      end if
    end if

    if (gtiType == GTI_IN_LIST) then
      do attBinNum = 1, numAttBins
        call andIntervals(attIntervals(attBinNum), gtis, andedIntervals)
        ! in ssclib/intervals_aux

        binnedAttWeights(attBinNum) = sum(andedIntervals%upper)&
           - sum(andedIntervals%lower)

        deallocate(andedIntervals)
      end do

    else ! gtiType == OPEN_GTI
      binnedAttWeights(1) = binnedAtt(1)%binEndTime - startTimeFirstBin
      do attBinNum = 2, numAttBins
        binnedAttWeights(attBinNum) = binnedAtt(attBinNum)%binEndTime&
          - binnedAtt(attBinNum-1)%binEndTime
      end do
    end if

  end subroutine calcBinnedAttGtis

  !---------------------------------------------------------------------
  subroutine openGtiWeightBinnedAtt(binnedAtt, startTimeFirstBin&
    , endsUnbounded, multiplyByGtis, binnedAttWeights)
  !---------------------------------------------------------------------

    implicit none

    type(attBinType), intent(in)  :: binnedAtt(:)
    real(double),     intent(in)  :: startTimeFirstBin
    logical,          intent(in)  :: endsUnbounded(2)
    logical,          intent(in)  :: multiplyByGtis
    real(single),     intent(out) :: binnedAttWeights(size(binnedAtt))

    integer :: numAttBins, attBinNum

    !-------------------------------------------------------------------
    ! Start

    numAttBins = size(binnedAtt)
  if (numAttBins <= 0) return

    if (.not. multiplyByGtis) then
      binnedAttWeights = 1.0
  return
    end if

if (any(endsUnbounded)) call error('',"")

    binnedAttWeights(1) = binnedAtt(1)%binEndTime - startTimeFirstBin
    do attBinNum = 2, numAttBins
      binnedAttWeights(attBinNum) = binnedAtt(attBinNum)%binEndTime&
        - binnedAtt(attBinNum-1)%binEndTime
    end do

  end subroutine openGtiWeightBinnedAtt

  !---------------------------------------------------------------------
  subroutine constructDummyBinnedAtt(templateSetName, binnedAtt&
  , startTimeFirstBin, endsUnbounded)
  !---------------------------------------------------------------------
  ! Note! The calling routine should deallocate the pointer binnedAtt after use.

    implicit none 

    character(*),     intent(in)  :: templateSetName
    type(attBinType), pointer     :: binnedAtt(:)
    real(double),     intent(out) :: startTimeFirstBin
    logical,          intent(out) :: endsUnbounded(2)

    type(DataSetT) :: set
    character(20) :: dateTimeStr
    real(double) ::&
      pntRaDeg,&
      pntDecDeg,&
      pntAposDeg

    !-------------------------------------------------------------------
    ! Start:

    allocate(binnedAtt(1))
    set = dataSet(templateSetName, READ)

    dateTimeStr = stringAttribute(set, "DATE-OBS")
    startTimeFirstBin = toTimeTag(dateTimeStr) ! in caloalutils/caloaldefs.f90

    pntRaDeg    = real64Attribute(set, "RA_PNT")
    pntDecDeg   = real64Attribute(set, "DEC_PNT")
    pntAposDeg  = real64Attribute(set, "PA_PNT")

    binnedAtt(1)%att%ra   = pntRaDeg   * PiOn180
    binnedAtt(1)%att%dec  = pntDecDeg  * PiOn180
    binnedAtt(1)%att%apos = pntAposDeg * PiOn180

    binnedAtt(1)%fracexp    = 1.0
    binnedAtt(1)%isGood     = .true.
    binnedAtt(1)%binEndTime = startTimeFirstBin

    endsUnbounded = .true.

    call release(set)

  end subroutine constructDummyBinnedAtt

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module binned_att
